// file      : xsd/cxx/tree/elements.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_TREE_ELEMENTS_HXX
#define CXX_TREE_ELEMENTS_HXX

#include <sstream>

#include <cult/containers/map.hxx>
#include <cult/containers/deque.hxx>
#include <cult/containers/vector.hxx>

#include <cxx/elements.hxx>

#include <cxx/tree/cli.hxx>


namespace CXX
{
  namespace Tree
  {
    struct Counts
    {
      Counts ()
          : global_types (0),
            global_elements (0),
            generated_global_elements (0),
            complexity_total (0)
      {
      }

      UnsignedLong global_types;
      UnsignedLong global_elements;
      UnsignedLong generated_global_elements;

      // Complexity value for each global type and generated global
      // element, in order.
      //
      Cult::Containers::Vector<UnsignedLong> complexity;
      UnsignedLong complexity_total;
    };

    struct InvalidCustomTypeMapping
    {
      InvalidCustomTypeMapping (String const& mapping,
                                String const& reason)
          : mapping_ (mapping), reason_ (reason)
      {
      }

      String const&
      mapping () const
      {
        return mapping_;
      }

      String const&
      reason () const
      {
        return reason_;
      }

    private:
      String mapping_;
      String reason_;
    };

    //
    //
    class Context: public CXX::Context
    {
    public:
      struct DirectCustomTypeMapInfo
      {
        DirectCustomTypeMapInfo (String const& t = L"",
                                 String const& b = L"")
            : type (t), base (b)
        {
        }

        String type;
        String base;
      };

      struct RegexCustomTypeMapInfo
      {
        RegexCustomTypeMapInfo (String const& p,
                                String const& t,
                                String const& b)
            : pat (p), type_sub (t), base_sub (b)
        {
        }

        String pat;
        String type_sub;
        String base_sub;
      };

      typedef
      Cult::Containers::Vector<RegexCustomTypeMapInfo>
      RegexCustomTypeMap;

      typedef
      Cult::Containers::Map<String, DirectCustomTypeMapInfo>
      DirectCustomTypeMap;

    public:
      Context (std::wostream& o,
               SemanticGraph::Schema& root,
               CLI::Options const& ops,
               Counts const& counts_,
               Boolean generate_xml_schema);

    protected:
      Context (Context& c);
      Context (Context& c, std::wostream& o);

      // Custom type mapping.
      //
    public:
      // Returns empty string if custom mapping is not required.
      //
      String
      custom_type (SemanticGraph::Type const&) const;

      // Returns true if custom mapping is required. name is
      // populated with the custom type name or empty if the
      // original name should be used.
      //
      Boolean
      custom_type (SemanticGraph::Type const&, String& name) const;

      // Returns true if this type has been renamed as part of the
      // customization process. If the function returns true, the
      // name string is populated with the new name or empty if
      // the type should not be generated at all.
      //
      Boolean
      renamed_type (SemanticGraph::Type const&, String& name) const;

    public:
      // Performs a number of processing steps, including forcing a new
      // line after 80 characters as well as "commentizing" the text by
      // adding '* ' after each newline.
      //
      Void
      write_annotation (SemanticGraph::Annotation&);

      // Escaped names.
      //
    public:
      static String const&
      etype (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("type");
      }

      static String const&
      etraits (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("traits");
      }

      static String const&
      econtainer (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("container");
      }

      static String const&
      econtainer (SemanticGraph::Any const& a)
      {
        return a.context ().get<String> ("container");
      }

      static String const&
      econtainer (SemanticGraph::AnyAttribute const& a)
      {
        return a.context ().get<String> ("container");
      }

      static String const&
      eiterator (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("iterator");
      }

      static String const&
      eiterator (SemanticGraph::Any const& a)
      {
        return a.context ().get<String> ("iterator");
      }

      static String const&
      eiterator (SemanticGraph::AnyAttribute const& a)
      {
        return a.context ().get<String> ("iterator");
      }

      static String const&
      econst_iterator (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("const-iterator");
      }

      static String const&
      econst_iterator (SemanticGraph::Any const& a)
      {
        return a.context ().get<String> ("const-iterator");
      }

      static String const&
      econst_iterator (SemanticGraph::AnyAttribute const& a)
      {
        return a.context ().get<String> ("const-iterator");
      }

      static String const&
      emember (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("member");
      }

      static String const&
      emember (SemanticGraph::Any const& a)
      {
        return a.context ().get<String> ("member");
      }

      static String const&
      emember (SemanticGraph::AnyAttribute const& a)
      {
        return a.context ().get<String> ("member");
      }

      static String const&
      edefault_value (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("default-value");
      }

      static String const&
      edefault_value_member (SemanticGraph::Member const& m)
      {
        return m.context ().get<String> ("default-value-member");
      }

      // Underlying enum value type.
      //
      static String const&
      evalue (SemanticGraph::Enumeration const& e)
      {
        return e.context ().get<String> ("value");
      }

      // dom_document
      //
      static Boolean
      edom_document_p (SemanticGraph::Complex const& c)
      {
        return c.context ().count ("dom-document");
      }

      static String const&
      edom_document (SemanticGraph::Complex const& c)
      {
        return c.context ().get<String> ("dom-document");
      }

      static Boolean
      edom_document_member_p (SemanticGraph::Complex const& c)
      {
        return c.context ().count ("dom-document-member");
      }

      static String const&
      edom_document_member (SemanticGraph::Complex const& c)
      {
        return c.context ().get<String> ("dom-document-member");
      }

    public:
      Void
      enter_ns_scope (String const& name)
      {
        ns_scope_stack.push_back (name);
        update_ns_scope ();
      }

      Void
      leave_ns_scope ()
      {
        ns_scope_stack.pop_back ();
        update_ns_scope ();
      }

    private:
      Void
      update_ns_scope ();

    private:
      // Write text that may contain characters that we will have
      // to escape (indicated by the rogue flag).
      //
      Void
      write_rogue_text (WideChar const* s, Size size, Boolean rogue);

    public:
      CLI::Options const& options;
      Counts const& counts;
      String& any_type;
      String& any_simple_type;
      String& flags_type;
      String& qname_type;
      String& parser_type;
      String& ostream_type;
      Boolean& generate_xml_schema;
      Boolean& doxygen;

      String& ns_scope;

      RegexCustomTypeMap& regex_custom_type_map;
      DirectCustomTypeMap& direct_custom_type_map;

    private:
      CLI::Options const options_;
      String any_type_;
      String any_simple_type_;
      String flags_type_;
      String qname_type_;
      String parser_type_;
      String ostream_type_;
      Boolean generate_xml_schema_;
      Boolean doxygen_;

      typedef
      Cult::Containers::Deque<String>
      NamespaceStack;

      typedef
      Cult::Containers::Deque<String>
      ScopeStack;

      String ns_scope_;

      NamespaceStack& ns_scope_stack;
      NamespaceStack ns_scope_stack_;

      RegexCustomTypeMap regex_custom_type_map_;
      DirectCustomTypeMap direct_custom_type_map_;
    };

    // Check whether this Schema type maps to a fundamental C++ type.
    //
    struct IsFundamentalType : Traversal::Fundamental::Byte,
                               Traversal::Fundamental::UnsignedByte,
                               Traversal::Fundamental::Short,
                               Traversal::Fundamental::UnsignedShort,
                               Traversal::Fundamental::Int,
                               Traversal::Fundamental::UnsignedInt,
                               Traversal::Fundamental::Long,
                               Traversal::Fundamental::UnsignedLong,
                               Traversal::Fundamental::Integer,
                               Traversal::Fundamental::NonPositiveInteger,
                               Traversal::Fundamental::NonNegativeInteger,
                               Traversal::Fundamental::PositiveInteger,
                               Traversal::Fundamental::NegativeInteger,

                               Traversal::Fundamental::Boolean,

                               Traversal::Fundamental::Float,
                               Traversal::Fundamental::Double,
                               Traversal::Fundamental::Decimal

    {
      IsFundamentalType (Boolean& r)
          : r_ (r)
      {
      }

      // Integral types.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Byte&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedByte&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Short&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedShort&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Int&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedInt&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Long&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedLong&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Integer& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonPositiveInteger& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonNegativeInteger& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::PositiveInteger& i)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NegativeInteger& i)
      {
        r_ = true;
      }

      // Boolean.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Boolean&)
      {
        r_ = true;
      }

      // Floats.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Float&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Double&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Decimal& d)
      {
        r_ = true;
      }

    private:
      Boolean& r_;
    };

    // Check whether this is a string-based type.
    //
    struct IsStringBasedType : Traversal::Complex,
                               Traversal::Fundamental::String,
                               Traversal::Fundamental::NormalizedString,
                               Traversal::Fundamental::Token,
                               Traversal::Fundamental::Name,
                               Traversal::Fundamental::NameToken,
                               Traversal::Fundamental::NCName,
                               Traversal::Fundamental::Language
    {
      IsStringBasedType (Boolean& r)
          : r_ (r)
      {
        *this >> inherits_ >> *this;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        inherits (c, inherits_);
      }

      // Strings.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::String&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NormalizedString&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Token&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameToken&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Name&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NCName&)
      {
        r_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Language&)
      {
        r_ = true;
      }

    private:
      Boolean& r_;
      Traversal::Inherits inherits_;
    };


    // Check whether this is a enumeration-based type.
    //
    struct IsEnumBasedType : Traversal::Complex
    {
      IsEnumBasedType (SemanticGraph::Enumeration*& e)
          : enum_ (e)
      {
        *this >> inherits_;

        inherits_ >> *this;
        inherits_ >> enum_;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        inherits (c, inherits_);
      }

    private:
      struct Enumeration: Traversal::Enumeration
      {
        Enumeration (SemanticGraph::Enumeration*& e)
            : e_ (e)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          if (e_ == 0)
            e_ = &e;
        }

      private:
        SemanticGraph::Enumeration*& e_;
      };


    private:
      Enumeration enum_;
      Traversal::Inherits inherits_;
    };


    //
    //
    struct MemberTypeName : protected Context,
                            Traversal::Type,
                            Traversal::List,
                            Traversal::Union,
                            Traversal::Complex,

                            Traversal::AnyType,
                            Traversal::AnySimpleType,

                            Traversal::Fundamental::Byte,
                            Traversal::Fundamental::UnsignedByte,
                            Traversal::Fundamental::Short,
                            Traversal::Fundamental::UnsignedShort,
                            Traversal::Fundamental::Int,
                            Traversal::Fundamental::UnsignedInt,
                            Traversal::Fundamental::Long,
                            Traversal::Fundamental::UnsignedLong,
                            Traversal::Fundamental::Integer,
                            Traversal::Fundamental::NonPositiveInteger,
                            Traversal::Fundamental::NonNegativeInteger,
                            Traversal::Fundamental::PositiveInteger,
                            Traversal::Fundamental::NegativeInteger,

                            Traversal::Fundamental::Boolean,

                            Traversal::Fundamental::Float,
                            Traversal::Fundamental::Double,
                            Traversal::Fundamental::Decimal,

                            Traversal::Fundamental::String,
                            Traversal::Fundamental::NormalizedString,
                            Traversal::Fundamental::Token,
                            Traversal::Fundamental::Name,
                            Traversal::Fundamental::NameToken,
                            Traversal::Fundamental::NameTokens,
                            Traversal::Fundamental::NCName,
                            Traversal::Fundamental::Language,

                            Traversal::Fundamental::QName,

                            Traversal::Fundamental::Id,
                            Traversal::Fundamental::IdRef,
                            Traversal::Fundamental::IdRefs,

                            Traversal::Fundamental::AnyURI,

                            Traversal::Fundamental::Base64Binary,
                            Traversal::Fundamental::HexBinary,

                            Traversal::Fundamental::Date,
                            Traversal::Fundamental::DateTime,
                            Traversal::Fundamental::Duration,
                            Traversal::Fundamental::Day,
                            Traversal::Fundamental::Month,
                            Traversal::Fundamental::MonthDay,
                            Traversal::Fundamental::Year,
                            Traversal::Fundamental::YearMonth,
                            Traversal::Fundamental::Time,

                            Traversal::Fundamental::Entity,
                            Traversal::Fundamental::Entities

    {
      MemberTypeName (Context& c)
          : Context (c)
      {
      }

      MemberTypeName (Context& c, std::wostream& o)
          : Context (c, o)
      {
      }

      virtual Void
      traverse (SemanticGraph::Type& t)
      {
        abort ();
      }

      virtual Void
      traverse (SemanticGraph::List& l)
      {
        os << fq_name (l);
      }

      virtual Void
      traverse (SemanticGraph::Union& u)
      {
        os << fq_name (u);
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        os << fq_name (c);
      }

      // anyType & anySimpleType.
      //
      virtual Void
      traverse (SemanticGraph::AnyType& i)
      {
        os << ns (i) << "::type";
      }

      virtual Void
      traverse (SemanticGraph::AnySimpleType& i)
      {
        os << ns (i) << "::simple_type";
      }

      // Integral types.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Byte& i)
      {
        os << ns (i) << "::byte";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedByte& i)
      {
        os << ns (i) << "::unsigned_byte";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Short& i)
      {
        os << ns (i) << "::short_";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedShort& i)
      {
        os << ns (i) << "::unsigned_short";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Int& i)
      {
        os << ns (i) << "::int_";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedInt& i)
      {
        os << ns (i) << "::unsigned_int";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Long& i)
      {
        os << ns (i) << "::long_";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedLong& i)
      {
        os << ns (i) << "::unsigned_long";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Integer& i)
      {
        os << ns (i) << "::integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonPositiveInteger& i)
      {
        os << ns (i) << "::non_positive_integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonNegativeInteger& i)
      {
        os << ns (i) << "::non_negative_integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::PositiveInteger& i)
      {
        os << ns (i) << "::positive_integer";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NegativeInteger& i)
      {
        os << ns (i) << "::negative_integer";
      }

      // Boolean.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Boolean& i)
      {
        os << ns (i) << "::boolean";
      }

      // Floats.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Float& i)
      {
        os << ns (i) << "::float_";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Double& i)
      {
        os << ns (i) << "::double_";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Decimal& d)
      {
        os << ns (d) << "::decimal";
      }

      // Strings.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::String& s)
      {
        os << ns (s) << "::string";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NormalizedString& s)
      {
        os << ns (s) << "::normalized_string";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Token& t)
      {
        os << ns (t) << "::token";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameToken& t)
      {
        os << ns (t) << "::nmtoken";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameTokens& t)
      {
        os << ns (t) << "::nmtokens";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Name& n)
      {
        os << ns (n) << "::name";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NCName& n)
      {
        os << ns (n) << "::ncname";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Language& l)
      {
        os << ns (l) << "::language";
      }


      // Qualified name.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::QName& n)
      {
        os << ns (n) << "::qname";
      }


      // ID/IDREF.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Id& i)
      {
        os << ns (i) << "::id";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRef& i)
      {
        if (i.named ())
        {
          // IDREF<anyType>
          //
          os << ns (i) << "::idref";
        }
        else
        {
          os << "::xsd::cxx::tree::idref< " <<
            type_name (i.argumented ().type ()) << ", " <<
            char_type << ", " <<
            xs_ns_name () << "::ncname >";
        }
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRefs& i)
      {
        if (i.named ())
        {
          // IDREFS<anyType>
          //
          os << ns (i) << "::idrefs";
        }
        else
        {
          os << "::xsd::cxx::tree::idrefs< " << char_type << ", " <<
            xs_ns_name () << "::simple_type, " <<
            "::xsd::cxx::tree::idref< " <<
            type_name (i.argumented ().type ()) << ", " <<
            char_type << ", " <<
            xs_ns_name () << "::ncname > >";
        }
      }

      // URI.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::AnyURI& u)
      {
        os << ns (u) << "::uri";
      }

      // Binary.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Base64Binary& b)
      {
        os << ns (b) << "::base64_binary";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::HexBinary& b)
      {
        os << ns (b) << "::hex_binary";
      }


      // Date/time.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Date& d)
      {
        os << ns (d) << "::date";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::DateTime& dt)
      {
        os << ns (dt) << "::date_time";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Duration& d)
      {
        os << ns (d) << "::duration";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Day& d)
      {
        os << ns (d) << "::day";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Month& m)
      {
        os << ns (m) << "::month";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::MonthDay& md)
      {
        os << ns (md) << "::month_day";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Year& y)
      {
        os << ns (y) << "::year";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::YearMonth& ym)
      {
        os << ns (ym) << "::year_month";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Time& t)
      {
        os << ns (t) << "::time";
      }

      // Entity.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Entity& e)
      {
        os << ns (e) << "::entity";
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Entities& e)
      {
        os << ns (e) << "::entities";
      }

    private:
      String
      ns (SemanticGraph::Type& t)
      {
        return ns_name (dynamic_cast<SemanticGraph::Namespace&> (t.scope ()));
      }

      // For idref/idrefs
      //
      String
      type_name (SemanticGraph::Type& t)
      {
        // This type is always named.
        //
        std::wostringstream o;

        MemberTypeName type (*this, o);
        type.dispatch (t);

        return o.str ();
      }
    };


    //
    //
    struct BaseTypeName : MemberTypeName
    {
      BaseTypeName (Context& c)
          : MemberTypeName (c)
      {
      }

      BaseTypeName (Context& c, std::wostream& o)
          : MemberTypeName (c, o)
      {
      }

      virtual Void
      fundamental_base (String const& type)
      {
        String ns (xs_ns_name ());

        os << "::xsd::cxx::tree::fundamental_base< " <<
          ns << "::" << type << ", " <<
          char_type << ", " <<
          ns << "::simple_type >";
      }

      // Integrals.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Byte&)
      {
        fundamental_base ("byte");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedByte&)
      {
        fundamental_base ("unsigned_byte");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Short&)
      {
        fundamental_base ("short_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedShort&)
      {
        fundamental_base ("unsigned_short");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Int&)
      {
        fundamental_base ("int_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedInt&)
      {
        fundamental_base ("unsigned_int");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Long&)
      {
        fundamental_base ("long_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::UnsignedLong&)
      {
        fundamental_base ("unsigned_long");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Integer& i)
      {
        fundamental_base ("integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonPositiveInteger& i)
      {
        fundamental_base ("non_positive_integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NonNegativeInteger& i)
      {
        fundamental_base ("non_negative_integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::PositiveInteger& i)
      {
        fundamental_base ("positive_integer");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NegativeInteger& i)
      {
        fundamental_base ("negative_integer");
      }

      // Boolean.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Boolean&)
      {
        fundamental_base ("boolean");
      }

      // Floats.
      //
      virtual Void
      traverse (SemanticGraph::Fundamental::Float&)
      {
        fundamental_base ("float_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Double&)
      {
        fundamental_base ("double_");
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Decimal& d)
      {
        fundamental_base ("decimal");
      }
    };


    // Initial value should be true.
    //
    struct IsSimpleType : Traversal::Complex,
                          Traversal::Member,
                          Traversal::Any,
                          Traversal::AnyAttribute
    {
      IsSimpleType (Boolean& v)
          : v_ (v)
      {
        *this >> names_ >> *this;
        *this >> inherits_ >> *this;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        names (c, names_);

        if (v_)
          inherits (c, inherits_);
      }

      virtual Void
      traverse (SemanticGraph::Member&)
      {
        v_ = false;
      }

      virtual Void
      traverse (SemanticGraph::Any&)
      {
        v_ = false;
      }

      virtual Void
      traverse (SemanticGraph::AnyAttribute&)
      {
        v_ = false;
      }

    private:
      Boolean& v_;
      Traversal::Names names_;
      Traversal::Inherits inherits_;
    };

    // Test whether we need to generate default c-tor. Note that we are not
    // interested in anyAttribute since it is always mapped to a sequence.
    //
    struct GenerateDefaultCtor: Traversal::Complex,
                                Traversal::Enumeration,
                                Traversal::Type,
                                Traversal::Member,
                                Traversal::Any,
                                protected virtual Context
    {
      // generate should initially be false.
      //
      GenerateDefaultCtor (Context&, Boolean& generate, Boolean no_base);

      virtual Void
      traverse (SemanticGraph::Complex&);

      virtual Void
      traverse (SemanticGraph::Type&);

      virtual Void
      traverse (SemanticGraph::Enumeration&);

      virtual Void
      traverse (SemanticGraph::Member&);

      virtual Void
      traverse (SemanticGraph::Any&);

    private:
      Boolean& generate_;
      Boolean no_base_;

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;
    };

    // Test whether we need to generate from-base c-tor.
    //
    struct GenerateFromBaseCtor: Traversal::Complex
    {
      // generate should initially be false.
      //
      GenerateFromBaseCtor (Context& c, Boolean& generate);

      Void
      traverse (SemanticGraph::Complex& c);

    private:
      // Note that we are not interested in anyAttribute since it is always
      // mapped to a sequence.
      //
      struct Traverser: Traversal::Complex,
                        Traversal::Member,
                        Traversal::Any,
                        protected virtual Context
      {
        Traverser (Context& c, Boolean& generate);

        virtual Void
        traverse (SemanticGraph::Complex&);

        virtual Void
        traverse (SemanticGraph::Member&);

        virtual Void
        traverse (SemanticGraph::Any&);

      private:
        Boolean& generate_;

      private:
        Traversal::Inherits inherits_;
        Traversal::Names names_;
      } traverser_;

      Traversal::Inherits inherits_;
    };


    // Immediate non-optional member. Note that AnyAttribute is always
    // mapped to a sequence.
    //
    struct FromBaseCtorArg : Traversal::Any,
                             Traversal::Member,
                             protected virtual Context
    {
      FromBaseCtorArg (Context& c, Boolean arg);

      virtual Void
      traverse (SemanticGraph::Any&);

      virtual Void
      traverse (SemanticGraph::Member&);

    private:
      Boolean arg_;
    };

    // List of all non-optional members and a simple base. Note that
    // AnyAttribute is always mapped to a sequence.
    //
    struct CtorArgs : Traversal::Complex,
                      Traversal::Enumeration,
                      Traversal::Type,
                      Traversal::Any,
                      Traversal::Member,
                      protected virtual Context
    {
      // The second version outputs the argument name and stores
      // in in the base_arg string.
      //
      CtorArgs (Context& c);
      CtorArgs (Context& c, String& base_arg);

      virtual Void
      traverse (SemanticGraph::Type&);

      virtual Void
      traverse (SemanticGraph::Enumeration&);

      virtual Void
      traverse (SemanticGraph::Any&);

      virtual Void
      traverse (SemanticGraph::Member&);

    private:
      String
      comma ();

    private:
      String base_;
      String* base_arg_;
      Boolean first_;

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;

      MemberTypeName member_name_;
    };


    // Check whether we need to generate c-tor without the base argument.
    //
    struct GenerateWithoutBaseCtor: Traversal::List,
                                    Traversal::Union,
                                    Traversal::Complex,
                                    Traversal::Enumeration,

                                    Traversal::AnyType,
                                    Traversal::AnySimpleType,

                                    Traversal::Fundamental::String,
                                    Traversal::Fundamental::NormalizedString,
                                    Traversal::Fundamental::Token,

                                    Traversal::Fundamental::NameTokens,
                                    Traversal::Fundamental::IdRefs,

                                    Traversal::Fundamental::Base64Binary,
                                    Traversal::Fundamental::HexBinary
    {
      // generate should initially be false.
      //
      GenerateWithoutBaseCtor (Boolean& generate)
          : generate_ (generate)
      {
        *this >> inherits_ >> *this;
      }

      virtual Void
      traverse (SemanticGraph::List&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Union&)
      {
        // No default initialization.
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        Complex::inherits (c);
      }

      virtual Void
      traverse (SemanticGraph::Enumeration&)
      {
        // No default initialization.
      }

      virtual Void
      traverse (SemanticGraph::AnyType&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::AnySimpleType&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::String&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NormalizedString&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Token&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::NameTokens&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::IdRefs&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::Base64Binary&)
      {
        generate_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Fundamental::HexBinary&)
      {
        generate_ = true;
      }

    private:
      Boolean& generate_;
      Traversal::Inherits inherits_;
    };


    // List of all non-optional members sans simple base. Note that
    // AnyAttribute is always mapped to a sequence.
    //
    struct CtorArgsWithoutBase: Traversal::Complex,
                                Traversal::Any,
                                Traversal::Member,
                                protected virtual Context
    {
      CtorArgsWithoutBase (Context& c, Boolean arg);

      virtual Void
      traverse (SemanticGraph::Any&);

      virtual Void
      traverse (SemanticGraph::Member&);

    private:
      String
      comma ();

    private:
      Boolean arg_;
      Boolean first_;

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;
    };

    //
    //
    struct GlobalElementBase
    {
      GlobalElementBase (Context& c)
          : ctx_ (c)
      {
      }

      Boolean
      generate_p (SemanticGraph::Element&);

      Boolean
      doc_root_p (SemanticGraph::Element&);

    private:
      Context& ctx_;
    };


    //
    //
    struct Namespace: CXX::Namespace,
                      GlobalElementBase,
                      CXX::Namespace::ScopeTracker
    {
      Namespace (Context&,
                 UnsignedLong first = 1,
                 UnsignedLong last = 0);

      virtual Void
      traverse (Type&);

    protected:
      virtual Void
      enter (Type&, String const& name, Boolean last);

      virtual Void
      leave ();

    protected:
      Context& ctx_;

    private:
      UnsignedLong first_;
      UnsignedLong last_;
      UnsignedLong count_;
    };
  }
}

#endif  // CXX_TREE_ELEMENTS_HXX
