// file      : xsd/cxx/tree/name-processor.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006-2007 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/tree/name-processor.hxx>
#include <cxx/tree/elements.hxx>

#include <cult/containers/set.hxx>
#include <cult/containers/map.hxx>

#include <iostream>

namespace CXX
{
  namespace Tree
  {
    namespace
    {
      //
      //
      typedef Cult::Containers::Set<String> NameSet;

      class Context: public Tree::Context
      {
      public:
        Context (CLI::Options const& options,
                 Counts const& counts,
                 Boolean generate_xml_schema,
                 SemanticGraph::Schema& root,
                 SemanticGraph::Path const& file)
            : Tree::Context (std::wcerr,
                             root,
                             options,
                             counts,
                             generate_xml_schema),
              schema_path_ (file),
              schema_path (schema_path_),
              ns_ (0),
              ns (ns_),
              global_type_names (global_type_names_),
              global_element_names (global_element_names_)
        {
        }

      protected:
        Context (Context& c)
            : Tree::Context (c),
              schema_path (c.schema_path),
              ns (c.ns),
              global_type_names (c.global_type_names),
              global_element_names (c.global_element_names)
        {
        }

      private:
        SemanticGraph::Path const schema_path_;
        SemanticGraph::Namespace* ns_;

        Cult::Containers::Map<String, NameSet> global_type_names_;
        Cult::Containers::Map<String, NameSet> global_element_names_;

      public:
        SemanticGraph::Path const& schema_path;
        SemanticGraph::Namespace*& ns;

        Cult::Containers::Map<String, NameSet>& global_type_names;
        Cult::Containers::Map<String, NameSet>& global_element_names;
      };

      String
      find_name (String const& base_name, NameSet& set)
      {
        String name (base_name);

        for (UnsignedLong i (1); set.find (name) != set.end (); ++i)
        {
          std::wostringstream os;
          os << i;
          name = base_name + os.str ();
        }

        set.insert (name);
        return name;
      }

      String
      find_name (SemanticGraph::Nameable& n, NameSet& set)
      {
        String r (find_name (Tree::Context::escape (n.name ()), set));
        n.context ().set ("name", r);
        return r;
      }

      //
      //
      struct Enumerator: Traversal::Enumerator
      {
        Enumerator (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          find_name (e, set_);
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct Enumeration: Traversal::Enumeration, Context
      {
        Enumeration (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          // Use processed name.
          //
          String name (e.context ().get<String> ("name"));

          // If renamed name is empty then we are not generating
          // anything for this type and name processing is not
          // required.
          //
          if (renamed_type (e, name) && !name)
            return;

          NameSet enum_set;
          enum_set.insert (name);

          Enumerator enumerator (enum_set);
          Traversal::Names names (enumerator);

          Traversal::Enumeration::names (e, names);

          // Assign name to the value type.
          //
          e.context ().set ("value", find_name ("value", enum_set));
        }
      };


      //
      //
      struct PrimaryMember: Traversal::Member
      {
        PrimaryMember (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (Tree::Context::skip (m))
            return;

          find_name (m, set_);
        }

      private:
        NameSet& set_;
      };

      //
      //
      struct DerivedMember: Traversal::Member
      {
        DerivedMember (NameSet& set)
            : set_ (set)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          if (Tree::Context::skip (m))
            return;

          String const& base (m.context ().get<String> ("name"));

          // Types.
          //
          m.context ().set ("type", find_name (base + L"_type", set_));
          m.context ().set ("traits", find_name (base + L"_traits", set_));

          if (Tree::Context::max (m) != 1)
          {
            m.context ().set (
              "container", find_name (base + L"_sequence", set_));

            m.context ().set (
              "iterator", find_name (base + L"_iterator", set_));

            m.context ().set (
              "const-iterator", find_name (base + L"_const_iterator", set_));
          }
          else if (Tree::Context::min (m) == 0 &&
                   (!m.is_a<SemanticGraph::Attribute> () || !m.default_ ()))
          {
            m.context ().set (
              "container", find_name (base + L"_optional", set_));
          }

          // Data member.
          //
          m.context ().set ("member", find_name (base + L"_", set_));

          // Default value.
          //
          if (m.default_ ())
          {
            Boolean simple (true);

            if (m.is_a<SemanticGraph::Element> ())
            {
              IsSimpleType test (simple);
              test.dispatch (m.type ());
            }

            if (simple)
            {
              m.context ().set (
                "default-value", find_name (base + L"_default_value", set_));

              m.context ().set (
                "default-value-member",
                find_name (base + L"_default_value_", set_));
            }
          }
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct Any: Traversal::Any, Traversal::AnyAttribute
      {
        Any (NameSet& set, Boolean& has_wildcard)
            : set_ (set), has_wildcard_ (has_wildcard)
        {
        }

        virtual Void
        traverse (SemanticGraph::Any& a)
        {
          String name (find_name (L"any", set_));

          a.context ().set ("name", name);

          // Types
          //
          if (Tree::Context::max (a) != 1)
          {
            a.context ().set (
              "container", find_name (name + L"_sequence", set_));

            a.context ().set (
              "iterator", find_name (name + L"_iterator", set_));

            a.context ().set (
              "const-iterator", find_name (name + L"_const_iterator", set_));
          }
          else if (Tree::Context::min (a) == 0)
          {
            a.context ().set (
              "container", find_name (name + L"_optional", set_));
          }

          // Data member.
          //
          a.context ().set ("member", find_name (name + L"_", set_));

          if (!has_wildcard_)
            has_wildcard_ = true;
        }

        virtual Void
        traverse (SemanticGraph::AnyAttribute& a)
        {
          String name (find_name (L"any_attribute", set_));

          a.context ().set ("name", name);

          // Types
          //
          a.context ().set (
            "container", find_name (name + L"_set", set_));

          a.context ().set (
            "iterator", find_name (name + L"_iterator", set_));

          a.context ().set (
            "const-iterator", find_name (name + L"_const_iterator", set_));

          // Data member.
          //
          a.context ().set ("member", find_name (name + L"_", set_));

          if (!has_wildcard_)
            has_wildcard_ = true;
        }

      private:
        NameSet& set_;
        Boolean& has_wildcard_;
      };

      //
      //
      struct Complex: Traversal::Complex, Context
      {
        Complex (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& c)
        {
          // Use processed name.
          //
          String name (c.context ().get<String> ("name"));

          // If renamed name is empty then we are not generating
          // anything for this type and name processing is not
          // required.
          //
          if (renamed_type (c, name) && !name)
            return;

          NameSet member_set;
          member_set.insert (name);

          // Add our base's members to the initial list.
          //
          if (c.inherits_p ())
          {
            // @@ What if this types name is the same as one of base's
            //    members?
            //
            SemanticGraph::Type& base (c.inherits ().base ());

            if (base.is_a<SemanticGraph::Complex> ())
            {
              if (!base.context ().count (
                    "cxx-tree-name-processor-member-set"))
              {
                dispatch (base);
              }

              NameSet const& base_set (
                base.context ().get<NameSet> (
                  "cxx-tree-name-processor-member-set"));

              member_set.insert (base_set.begin (), base_set.end ());
            }
          }

          // First assign the "primary" names.
          //
          {
            PrimaryMember member (member_set);
            Traversal::Names names (member);

            Complex::names (c, names);
          }

          // Derived names for members.
          //
          {
            DerivedMember member (member_set);
            Traversal::Names names (member);

            Complex::names (c, names);
          }

          // Names for wildcards.
          //
          if (options.value<CLI::generate_wildcard> ())
          {
            Boolean has_wildcard (false);
            Any any (member_set, has_wildcard);
            Traversal::Names names (any);
            Complex::names (c, names);

            // Assign names for dom_document.
            //
            if (has_wildcard)
            {
              // Check if we already have dom_document down inheritance
              // hierarchy.
              //
              for (SemanticGraph::Complex* p (&c); p->inherits_p ();)
              {
                if (SemanticGraph::Complex* base =
                    dynamic_cast<SemanticGraph::Complex*> (
                      &p->inherits ().base ()))
                {
                  if (base->context ().count ("dom-document"))
                  {
                    c.context ().set (
                      "dom-document",
                      base->context ().get<String> ("dom-document"));
                    break;
                  }

                  p = base;
                }
                else
                  break;
              }

              // If not, set up the names.
              //
              if (!c.context ().count ("dom-document"))
              {
                c.context ().set (
                  "dom-document",
                  find_name ("dom_document", member_set));

                c.context ().set (
                  "dom-document-member",
                  find_name ("dom_document_", member_set));
              }
            }
          }

          // @@ I am not removing this at the moment.
          //
          c.context ().set ("cxx-tree-name-processor-member-set", member_set);
        }
      };


      //
      //
      struct GlobalType: Traversal::Type, Context
      {
        GlobalType (Context& c, NameSet& set)
            : Context (c), set_ (set)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          String name (find_name (t, set_));

          // Also add renamed name if any.
          //
          if (renamed_type (t, name) && name)
            set_.insert (name);
        }

      private:
        NameSet& set_;
      };


      //
      //
      struct GlobalElement: Traversal::Element,
                            GlobalElementBase,
                            Context
      {
        GlobalElement (Context& c,
                       NameSet const& type_set,
                       NameSet& element_set)
            : GlobalElementBase (c),
              Context (c),
              type_set_ (type_set),
              element_set_ (element_set)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          // First we need to figure out if we need to process this
          // global element.
          //
          if (!generate_p (e))
            return;

          // Make sure the name is unique among global elements and
          // does not collide with a global type name.
          //
          String base (Tree::Context::escape (e.name ()));
          String name (base);

          // If we are conflicting with a type name let's first try to
          // simply append an underscore and only resort to ugly names
          // like name1, etc. if this fails.
          //
          if (type_set_.find (name) != type_set_.end ())
            name += L"_";

          for (UnsignedLong i (1);
               element_set_.find (name) != element_set_.end () ||
                 type_set_.find (name) != type_set_.end ();
               ++i)
          {
            std::wostringstream os;
            os << i;
            name = base + os.str ();
          }

          //if (name != base)
          //  os << "renamed '" << base << "' to '" << name << "'" << endl;

          element_set_.insert (name);
          e.context ().set ("name", name);
        }

        /*
      private:
        SemanticGraph::Type*
        find (SemanticGraph::Schema& schema, String const& name)
        {
          using SemanticGraph::Type;
          using SemanticGraph::Scope;
          using SemanticGraph::Namespace;

          String ns_name (ns->name ());

          // Get all namespaces across include/import hierarchy with
          // our namespace name.
          //
          // @@ using root schema here.
          //
          Scope::NamesIteratorPair nip (schema.find (ns_name));

          for (; nip.first != nip.second; ++nip.first)
          {
            Namespace& ns (dynamic_cast<Namespace&> (nip.first->named ()));

            Scope::NamesIteratorPair types (ns.find (name));

            for (; types.first != types.second; ++types.first)
            {
              if (Type* t = dynamic_cast<Type*> (&types.first->named ()))
              {
                return t;
              }
            }
          }

          return 0;
        }
        */

      private:
        NameSet const& type_set_;
        NameSet& element_set_;
      };

      struct NamespacePassOne: Traversal::Namespace, Context
      {
        NamespacePassOne (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& ns)
        {
          NameSet& type_set (global_type_names[ns.name ()]);

          GlobalType type (*this, type_set);
          Traversal::Names names (type);

          Context::ns = &ns;

          Traversal::Namespace::names (ns, names);
          Traversal::Namespace::names (ns);

          Context::ns = 0;
        }
      };


      struct NamespacePassTwo: Traversal::Namespace, Context
      {
        NamespacePassTwo (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& ns)
        {
          String const& name (ns.name ());

          NameSet const& type_set (global_type_names[name]);
          NameSet& element_set (global_element_names[name]);

          GlobalElement element (*this, type_set, element_set);
          Traversal::Names names (element);

          Context::ns = &ns;

          Traversal::Namespace::names (ns, names);

          Context::ns = 0;
        }
      };


      // Go into implied/included/imported schemas while making sure
      // we don't process the same stuff more than once.
      //
      struct UsesPassOne: Traversal::Uses
      {
        virtual Void
        traverse (Type& u)
        {
          SemanticGraph::Schema& s (u.schema ());

          if (!s.context ().count ("cxx-tree-name-processor-pass-1"))
          {
            s.context ().set ("cxx-tree-name-processor-pass-1", true);
            Traversal::Uses::traverse (u);
          }
        }
      };

      struct UsesPassTwo: Traversal::Uses
      {
        virtual Void
        traverse (Type& u)
        {
          SemanticGraph::Schema& s (u.schema ());

          if (!s.context ().count ("cxx-tree-name-processor-pass-2"))
          {
            s.context ().set ("cxx-tree-name-processor-pass-2", true);
            Traversal::Uses::traverse (u);
          }
        }
      };

      Void
      process_impl (CLI::Options const& ops,
                    SemanticGraph::Schema& tu,
                    SemanticGraph::Path const& file)
      {
        Counts counts;
        Context ctx (ops, counts, false, tu, file);

        // Pass one - assign names to global types. This pass cannot
        // be combined with pass two because of possible recursive
        // schema inclusions.
        //
        {
          Traversal::Schema schema;
          UsesPassOne uses;

          schema >> uses >> schema;

          Traversal::Names schema_names;
          NamespacePassOne ns (ctx);

          schema >> schema_names >> ns;

          // Some twisted schemas do recusive self-inclusion.
          //
          tu.context ().set ("cxx-tree-name-processor-pass-1", true);

          schema.dispatch (tu);
        }

        // Pass two - assign names inside complex/enum types. Here
        // we don't need to go into included/imported schemas.
        //
        {
          Traversal::Schema schema;
          Traversal::Sources sources;

          schema >> sources >> schema;

          Traversal::Names schema_names;
          Traversal::Namespace ns;
          Traversal::Names ns_names;

          schema >> schema_names >> ns >> ns_names;

          Complex complex (ctx);
          Enumeration enumeration (ctx);

          ns_names >> complex;
          ns_names >> enumeration;

          schema.dispatch (tu);
        }

        // Pass three - assign names to global elements.
        //
        {
          Traversal::Schema schema;
          UsesPassTwo uses;

          schema >> uses >> schema;

          Traversal::Names schema_names;
          NamespacePassTwo ns (ctx);

          schema >> schema_names >> ns;

          // Some twisted schemas do recusive self-inclusion.
          //
          tu.context ().set ("cxx-tree-name-processor-pass-2", true);

          schema.dispatch (tu);
        }
      }
    }

    Void NameProcessor::
    process (CLI::Options const& ops,
             SemanticGraph::Schema& tu,
             SemanticGraph::Path const& file)
    {
      process_impl (ops, tu, file);
    }
  }
}
